'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var react = require('react');
var proxyCompare = require('proxy-compare');
var useSubscription = require('use-subscription');

/*
export {
  unstable_createMutableSource as createMutableSource,
  unstable_useMutableSource as useMutableSource,
} from 'react'
*/
var TARGET = Symbol();
var createMutableSource = function createMutableSource(target, _getVersion) {
  var _ref;

  return _ref = {}, _ref[TARGET] = target, _ref;
};
var useMutableSource = function useMutableSource(source, getSnapshot, _subscribe) {
  var subscription = react.useMemo(function () {
    return {
      getCurrentValue: function getCurrentValue() {
        return getSnapshot(source[TARGET]);
      },
      subscribe: function subscribe(callback) {
        return _subscribe(source[TARGET], callback);
      }
    };
  }, [source, getSnapshot, _subscribe]);
  return useSubscription.useSubscription(subscription);
};

var VERSION = Symbol();
var LISTENERS = Symbol();
var SNAPSHOT = Symbol();

var isObject = function isObject(x) {
  return typeof x === 'object' && x !== null;
};

var proxyCache = new WeakMap();
var globalVersion = 0;
var snapshotCache = new WeakMap();
var proxy = function proxy(initialObject) {
  var _initialObject$constr;

  if (initialObject === void 0) {
    initialObject = {};
  }

  var version = globalVersion;
  var pendingVersion = version;
  var listeners = new Set();

  var notifyUpdate = function notifyUpdate(nextVersion) {
    if (!nextVersion) {
      nextVersion = ++globalVersion;
    }

    if (nextVersion > pendingVersion) {
      pendingVersion = nextVersion;
      Promise.resolve().then(function () {
        if (pendingVersion > version) {
          version = pendingVersion;
          listeners.forEach(function (listener) {
            return listener(version);
          });
        }
      });
    }
  };

  var emptyCopy = Array.isArray(initialObject) ? [] : Object.create(((_initialObject$constr = initialObject.constructor) == null ? void 0 : _initialObject$constr.prototype) || null);
  var p = new Proxy(emptyCopy, {
    get: function get(target, prop, receiver) {
      if (prop === VERSION) {
        return version;
      }

      if (prop === LISTENERS) {
        return listeners;
      }

      if (prop === SNAPSHOT) {
        var cache = snapshotCache.get(receiver);

        if (cache && cache.version === version) {
          return cache.snapshot;
        }

        var _snapshot = Object.create(target.constructor.prototype);

        snapshotCache.set(receiver, {
          version: version,
          snapshot: _snapshot
        });
        Reflect.ownKeys(target).forEach(function (key) {
          var value = target[key];

          if (!isObject(value)) {
            _snapshot[key] = value;
          } else if (value instanceof Promise) {
            Object.defineProperty(_snapshot, key, {
              get: function get() {
                throw value;
              }
            });
          } else {
            _snapshot[key] = value[SNAPSHOT];
          }
        });

        if (typeof process === 'object' && process.env.NODE_ENV !== 'production') {
          Object.freeze(_snapshot);
        }

        return _snapshot;
      }

      return target[prop];
    },
    deleteProperty: function deleteProperty(target, prop) {
      var value = target[prop];
      var childListeners = isObject(value) && value[LISTENERS];

      if (childListeners) {
        childListeners.delete(notifyUpdate);
      }

      delete target[prop];
      notifyUpdate();
      return true;
    },
    set: function set(target, prop, value, receiver) {
      var childListeners = isObject(target[prop]) && target[prop][LISTENERS];

      if (childListeners) {
        childListeners.delete(notifyUpdate);
      }

      if (!isObject(value)) {
        target[prop] = value;
      } else if (value instanceof Promise) {
        target[prop] = value.then(function (v) {
          receiver[prop] = v;
        });
      } else {
        value = proxyCompare.getUntrackedObject(value) || value;

        if (value[LISTENERS]) {
          target[prop] = value;
        } else if (proxyCache.has(value)) {
          target[prop] = proxyCache.get(value);
        } else {
          target[prop] = proxy(value);
        }

        target[prop][LISTENERS].add(notifyUpdate);
      }

      notifyUpdate();
      return true;
    }
  });
  proxyCache.set(initialObject, p);
  Reflect.ownKeys(initialObject).forEach(function (key) {
    p[key] = initialObject[key];
  });
  return p;
};
var subscribe = function subscribe(p, callback) {
  p[LISTENERS].add(callback);
  return function () {
    p[LISTENERS].delete(callback);
  };
};
var snapshot = function snapshot(p) {
  return p[SNAPSHOT];
};

var mutableSourceCache = new WeakMap();

var getMutableSource = function getMutableSource(p) {
  if (!mutableSourceCache.has(p)) {
    mutableSourceCache.set(p, createMutableSource(p));
  }

  return mutableSourceCache.get(p);
};

var useProxy = function useProxy(p) {
  var affected = new WeakMap();
  var lastAffected = react.useRef();
  react.useEffect(function () {
    lastAffected.current = affected;
  });
  var getChangedSnapshot = react.useMemo(function () {
    var prevSnapshot = null;
    var deepChangedCache = new WeakMap();
    return function (p) {
      var nextSnapshot = snapshot(p);

      try {
        if (prevSnapshot !== null && lastAffected.current && !proxyCompare.isDeepChanged(prevSnapshot, nextSnapshot, lastAffected.current, deepChangedCache)) {
          // not changed
          return prevSnapshot;
        }
      } catch (e) {// ignore and return new nextSnapshot
      }

      return prevSnapshot = nextSnapshot;
    };
  }, []);
  var currSnapshot = useMutableSource(getMutableSource(p), getChangedSnapshot, subscribe);
  var proxyCache = react.useMemo(function () {
    return new WeakMap();
  }, []); // per-hook proxyCache

  return proxyCompare.createDeepProxy(currSnapshot, affected, proxyCache);
};

exports.proxy = proxy;
exports.snapshot = snapshot;
exports.subscribe = subscribe;
exports.useProxy = useProxy;
